@@ -1,5 +1,5 @@
 /* grep.c - main driver file for grep.
-   Copyright (C) 1992, 1997-2002, 2004-2012 Free Software Foundation, Inc.
+   Copyright (C) 1992, 1997-2002, 2004-2013 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -44,7 +44,6 @@
 #include "progname.h"
 #include "propername.h"
 #include "quote.h"
-#include "stat-size.h"
 #include "version-etc.h"
 #include "xalloc.h"
 #include "xstrtol.h"
@@ -53,8 +52,6 @@
 #define SEP_CHAR_REJECTED '-'
 #define SEP_STR_GROUP    "--"
 
-#define STREQ(a, b) (strcmp (a, b) == 0)
-
 #define AUTHORS \
   proper_name ("Mike Haertel"), \
   _("others, see <http://git.sv.gnu.org/cgit/grep.git/tree/AUTHORS>")
@@ -443,9 +440,6 @@ clean_up_stdout (void)
 static int
 file_is_binary (char const *buf, size_t bufsize, int fd, struct stat const *st)
 {
-  #ifndef HAVE_STRUCT_STAT_ST_BLOCKS
-  enum { HAVE_STRUCT_STAT_ST_BLOCKS = 0 };
-  #endif
   #ifndef SEEK_HOLE
   enum { SEEK_HOLE = SEEK_END };
   #endif
@@ -461,8 +455,7 @@ file_is_binary (char const *buf, size_t bufsize, int fd, struct stat const *st)
     return 1;
 
   /* If the file has holes, it must contain a null byte somewhere.  */
-  if ((HAVE_STRUCT_STAT_ST_BLOCKS || SEEK_HOLE != SEEK_END)
-      && usable_st_size (st))
+  if (SEEK_HOLE != SEEK_END && usable_st_size (st))
     {
       off_t cur = bufsize;
       if (O_BINARY || fd == STDIN_FILENO)
@@ -472,28 +465,14 @@ file_is_binary (char const *buf, size_t bufsize, int fd, struct stat const *st)
             return 0;
         }
 
-      /* If the file has fewer blocks than would be needed to
-         represent its data, then it must have at least one hole.  */
-      if (HAVE_STRUCT_STAT_ST_BLOCKS)
-        {
-          off_t nonzeros_needed = st->st_size - cur + bufsize;
-          off_t full_blocks = nonzeros_needed / ST_NBLOCKSIZE;
-          int partial_block = 0 < nonzeros_needed % ST_NBLOCKSIZE;
-          if (ST_NBLOCKS (*st) < full_blocks + partial_block)
-            return 1;
-        }
-
       /* Look for a hole after the current location.  */
-      if (SEEK_HOLE != SEEK_END)
+      off_t hole_start = lseek (fd, cur, SEEK_HOLE);
+      if (0 <= hole_start)
         {
-          off_t hole_start = lseek (fd, cur, SEEK_HOLE);
-          if (0 <= hole_start)
-            {
-              if (lseek (fd, cur, SEEK_SET) < 0)
-                suppressible_error (filename, errno);
-              if (hole_start < st->st_size)
-                return 1;
-            }
+          if (lseek (fd, cur, SEEK_SET) < 0)
+            suppressible_error (filename, errno);
+          if (hole_start < st->st_size)
+            return 1;
         }
     }
 
@@ -1219,8 +1198,10 @@ grep (int fd, struct stat const *st)
          the buffer, 0 means there is no incomplete last line).  */
       oldc = beg[-1];
       beg[-1] = eol;
-      for (lim = buflim; lim[-1] != eol; lim--)
-        continue;
+      /* FIXME: use rawmemrchr if/when it exists, since we have ensured
+         that this use of memrchr is guaranteed never to return NULL.  */
+      lim = memrchr (beg - 1, eol, buflim - beg + 1);
+      ++lim;
       beg[-1] = oldc;
       if (lim == beg)
         lim = beg - residue;
@@ -1765,16 +1746,20 @@ static int
 get_nondigit_option (int argc, char *const *argv, intmax_t *default_context)
 {
   static int prev_digit_optind = -1;
-  int opt, this_digit_optind, was_digit;
+  int this_digit_optind, was_digit;
   char buf[INT_BUFSIZE_BOUND (intmax_t) + 4];
   char *p = buf;
+  int opt;
 
   was_digit = 0;
   this_digit_optind = optind;
-  while (opt = getopt_long (argc, (char **) argv, short_options, long_options,
-                            NULL),
-         '0' <= opt && opt <= '9')
+  while (1)
     {
+      opt = getopt_long (argc, (char **) argv, short_options,
+                         long_options, NULL);
+      if ( ! ('0' <= opt && opt <= '9'))
+        break;
+
       if (prev_digit_optind != this_digit_optind || !was_digit)
         {
           /* Reset to start another context length argument.  */
@@ -1879,6 +1864,7 @@ main (int argc, char **argv)
   size_t cc;
   int opt, status, prepended;
   int prev_optind, last_recursive;
+  int fread_errno;
   intmax_t default_context;
   FILE *fp;
   exit_failure = EXIT_TROUBLE;
@@ -2028,13 +2014,15 @@ main (int argc, char **argv)
           ;
         keys = xrealloc (keys, keyalloc);
         oldcc = keycc;
-        while (!feof (fp)
-               && (cc = fread (keys + keycc, 1, keyalloc - 1 - keycc, fp)) > 0)
+        while ((cc = fread (keys + keycc, 1, keyalloc - 1 - keycc, fp)) != 0)
           {
             keycc += cc;
             if (keycc == keyalloc - 1)
               keys = x2nrealloc (keys, &keyalloc, sizeof *keys);
           }
+        fread_errno = errno;
+        if (ferror (fp))
+          error (EXIT_TROUBLE, fread_errno, "%s", optarg);
         if (fp != stdin)
           fclose (fp);
         /* Append final newline if file ended in non-newline. */
@@ -2203,7 +2191,7 @@ main (int argc, char **argv)
     color_option = isatty (STDOUT_FILENO) && should_colorize ();
   init_colorize ();
 
-  /* POSIX.2 says that -q overrides -l, which in turn overrides the
+  /* POSIX says that -q overrides -l, which in turn overrides the
      other output options.  */
   if (exit_on_match)
     list_files = 0;
